from inspect import getargvalues, signature, _empty
from os import getcwd
from re import compile
from sys import settrace, path
from time import perf_counter


class Tracer:
    def __init__(self,
        label       = 'tracing',
        full_path   = False,
        show_types  = False,
        check_types = False,
        show_times  = False,
        stdlib      = False,
        args_format = 'str',
        dest        = None,
        buffered    = True,
        echo        = True,
        quiet       = True,
    ):
        self.label = label
        self.level = 0
        self.full_path = full_path
        self.traces = list()
        self.show_types = show_types
        self.check_types = check_types
        self.ignored_paths = set()
        self.ignored_funcs = set()
        self.truncated = list()
        self.timers = dict()
        self.show_times = show_times
        self.args_format = args_format
        self.dest = dest or getcwd()
        self.buffered = buffered
        self.echo = echo
        self.quiet = quiet
        self.stdlib = stdlib
        self.setup()

    def write(self, line):
        with open(f'{self.dest}/{self.label}.trace', 'a') as out:
            out.write(f'{line}\n')

    def setup(self):
        std_paths = [p for p in path if 'python' in p]
        if not self.stdlib:
            for p in std_paths:
                self.ignore_path(p)
            self.ignore_path('_bootstr')
        if not self.full_path:
            self.truncated.extend(std_paths)
        self.ignore_func(r'Tracer\..*')
        self.truncate_path(getcwd())
        self._exc_ = False

    def truncate(self, path):
        if not self.full_path:
            for pref in self.truncated:
                if path.startswith(pref):
                    path = path[len(pref)+1:]
        return path[:-3]

    def ignore_path(self, path):
        return self.ignored_paths.add(compile(path).search)

    def ignored_path(self, path):
        return any(ignore(path) for ignore in self.ignored_paths)

    def ignore_func(self, func):
        return self.ignored_funcs.add(compile(func).search)

    def ignored_func(self, func):
        return any(ignore(func) for ignore in self.ignored_funcs)

    def truncate_path(self, prefix):
        self.truncated.append(prefix)

    def start_timer(self):
        if self.show_times:
            self.timers[self.level] = perf_counter()

    def stop_timer(self):
        if self.show_times:
            self.timers[self.level] = perf_counter() - self.timers[self.level]

    @property
    def elapsed(self):
        if self.show_times:
            elapsed = self.timers[self.level]
            if elapsed < 10:
                elapsed = int(elapsed * 1000)
                return f' [{elapsed} ms]' if elapsed > 0 else ''
            else:
                return f' [{round(elapsed)} s]'
        else:
            return ''

    def log(self, msg):
        tab = '    ' * self.level
        msg = f'{tab}{msg}'
        if self.buffered:
            self.traces.append(msg)
        else:
            self.write(msg)
        if self.echo:
            print(msg)

    def log_exc(self, exc):
        if not self.quiet:
            msg = f'! {exc} !'
            self.log(msg)

    def format_args(self, args):
        if self.args_format == 'repr':
            formatter = repr
        elif self.args_format == 'str':
            formatter = str
        else:
            raise ValueError(f'Invalid args_formatter : {self.args_format}')

        items = dict()
        for k in args:
            try:
                v = formatter(args[k])
            except Exception as x:
                self.log_exc(x)
                v = ' ! '
            items[k] = v
        return items.items()

    def trace(self, file, line, func, args):
        if isinstance(args, tuple) and args[0] in ('>>>', '!!!'):
            if args[0] == '>>>':
                val = args[1]
                if self.show_types:
                    args = f' -> {type(val).__name__}: {val}{self.elapsed}'
                else:
                    args = f' -> {val}{self.elapsed}'
            elif args[0] == '!!!':
                args = f' ! {args[1].__name__} ! {args[2]}'
        else:
            if self.show_types:
                args = f"({', '.join(f'{k}:{type(v).__name__} = {v}' for k, v in self.format_args(args))})"
            else:
                args = f"({', '.join(f'{k}={v}' for k, v in self.format_args(args))})"
        self.log(f'{file}:{line} {func}{args}')


    def ann_err(self, name, expected, actual):
        if name == '_return_':
            msg = f'! "should return "{expected.__name__}" not "{actual.__name__}" !'
        else:
            msg = f'! arg "{name}" should be "{expected.__name__}" not "{actual.__name__}" !'
        self.log(msg)

    def get_func_args(self, frame):
        arginfo = getargvalues(frame)
        return {
            k:arginfo.locals[k]
            for k in arginfo.args
            if k != 'self'
        }

    def get_sig(self, frame):
        fct = frame.f_code.co_name
        if fct.endswith('comp>'): return
        slf = frame.f_locals.get('self')
        if slf:
            func = getattr(slf, fct)
        else:
            func = frame.f_globals[fct]
        return signature(func)

    def check_sig(self, frame, args):
        if self.check_types:
            sig = self.get_sig(frame)
            if not sig or not sig.parameters: return
            if isinstance(args, tuple) and args[0] == '>>>':
                a = sig.return_annotation
                if not a is _empty and not isinstance(args[1], a):
                    self.ann_err('_return_', a, type(args[1]))
            else:
                for n, p in sig.parameters.items():
                    a = p.annotation
                    if a is _empty: continue
                    if isinstance(args, dict) and not isinstance(args[n], a):
                        self.ann_err(n, a, type(args[n]))

    def get_func_name(self, frame):
        fct = frame.f_code.co_name
        slf = frame.f_locals.get('self')
        try:
            repr(slf) # check handlability
        except Exception:
            slf = None
        if slf:
            try:
                func = getattr(slf, fct)
            except Exception:
                func = None
        else:
            func = frame.f_globals.get(fct)
        try:
            return func.__qualname__
        except Exception:
            return fct

    def get_trace(self, frame, arg, lvl=0):
        path = frame.f_code.co_filename
        if self.ignored_path(path): return False
        func = self.get_func_name(frame)
        if self.ignored_func(func): return False
        self.level += lvl
        self.check_sig(frame, arg)
        path = self.truncate(path)
        line = frame.f_lineno
        self.trace(path, line, func, arg)
        return path, line, func, arg

    def trace_exception(self, frame, typ, val, tbk):
        ret = self.get_trace(frame, ('!!!', typ, val))
        if ret:
            self._exc_ = True

    def trace_return(self, frame, val):
        self.stop_timer()
        if self._exc_:
            self.level-=1
            self._exc_ = False
            return
        if self.get_trace(frame, ('>>>', val)):
            self.level-=1

    def trace_call(self, frame, event, arg):
        args = self.get_func_args(frame)
        if self.get_trace(frame, args, 1):
            self.start_timer()
        return self.trace_event

    def trace_event(self, frame, event, arg):
        if event == 'call':
            return self.trace_call

        if event == 'return':
            self.trace_return(frame, arg)

        if event == 'exception':
            self.trace_exception(frame, *arg)

    def dump(self):
        with open(f'{self.dest}/{self.label}.trace', 'w') as out:
            out.write('\n'.join(self.traces))

    def __enter__(self):
        if not self.buffered:
            with open(f'{self.dest}/{self.label}.trace', 'w') as out: pass
        self.log(f'{self.label} >>>')
        settrace(self.trace_event)

    def __exit__(self, typ, val, tbk):
        settrace(None)
        if self.level == 0:
            self.log(f'{self.label} <<<')
        else:
            self.log(f'{self.label} !!!')
        if self.buffered:
            self.dump()
        return self.quiet
